#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "endian.h"
#include "lz77.h"

#include "lzpt_io.h"


#define	LZPT_MAGIC_OFS		0
#define	LZPT_MAGIC	"TPZL"
#define	LZPT_MAGIC_LEN	(sizeof(LZPT_MAGIC)-1)

#define	LZPT_TOC_OFFSET_OFS	8
#define	LZPT_TOC_SIZE_OFS	12

#define	LZPT_BLOCK_ALIGN_MASK	(sizeof(unsigned int)-1)

#define	LZPT_MAX_DECOMP_BLOCK_SIZE	65536

int
lzpt_read_toc(FILE *fh_in, unsigned char **pp_toc, int *pnum_entries)
{
	unsigned char	iobuf[512];
	unsigned int	toc_offset, toc_size, toc_nentries;
	unsigned char	*p_toc = NULL;

	fseek(fh_in, LZPT_MAGIC_OFS, SEEK_SET);
	if ((fread(iobuf, 1, LZPT_MAGIC_LEN, fh_in) != LZPT_MAGIC_LEN) ||
		memcmp((void *)iobuf, (void *)LZPT_MAGIC, LZPT_MAGIC_LEN)) {
		fprintf(stderr, "lzpt_read_toc(): Bad magic!\n");
		return 1;
	}

	fseek(fh_in, LZPT_TOC_OFFSET_OFS, SEEK_SET);
	fread(iobuf, sizeof(unsigned int), 1, fh_in);
	toc_offset = readLE32(iobuf);

	fseek(fh_in, LZPT_TOC_SIZE_OFS, SEEK_SET);
	fread(iobuf, sizeof(unsigned int), 1, fh_in);
	toc_size = readLE32(iobuf);

	toc_nentries = (toc_size/(2 * sizeof(unsigned int)));

	if (!(p_toc = (unsigned char *)malloc(toc_size))) {
		fprintf(stderr, "lzpt_read_toc(): Error allocating TOC!\n");
		return 1;
	}

	memset(p_toc, 0, toc_size);
	fseek(fh_in, toc_offset, SEEK_SET);

	if (fread(p_toc, 1, toc_size, fh_in) != toc_size) {
		fprintf(stderr, "lzpt_read_toc(): Error reading TOC!\n");
		return 1;
	}

	*pp_toc = p_toc;
	*pnum_entries = toc_nentries;

	return 0;
}

int
lzpt_read_block(FILE *fh_in, int block_num, unsigned char *p_toc, size_t toc_entries, unsigned char **pp_block_data, size_t *psz_block)
{
	unsigned int	toc_offset, entry_offset, entry_size;
	unsigned char	*p_block;
	unsigned int	nread;

	*pp_block_data = NULL;
	*psz_block = 0;

	if (block_num >= toc_entries) {
		return 1;
	}
	toc_offset = (block_num * (sizeof(unsigned int) * 2));

	entry_offset = readLE32(p_toc+toc_offset);
	entry_size = readLE32(p_toc+toc_offset+4);

	if (!(p_block = malloc(entry_size))) {
		fprintf(stderr, "lzpt_read_block(): allocation error!\n");
		return 1;
	}
	fseek(fh_in, entry_offset, SEEK_SET);
	nread = fread(p_block, 1, entry_size, fh_in);

	if (nread != entry_size) {
		// special case: last block in file may be smaller by align size
		if ((block_num == (toc_entries-1)) && ((entry_size - nread) <= LZPT_BLOCK_ALIGN_MASK)) {
			// OK
		} else {
			fprintf(stderr, "lzpt_read_block(): block read error!\n");
			free((void *)p_block);
			return 1;
		}
	}
	*pp_block_data = p_block;
	*psz_block = nread;

	return 0;
}

int
lzpt_decompress_block(unsigned char *p_block_in, size_t sz_block_in, unsigned char **pp_block_out, size_t *psz_block_out)
{
	unsigned char	*p_block = NULL;
	int	remain_in, remain_out;
	int	this_declen;
	unsigned char	*p_in, *p_out, *next_in;
	size_t	total_decomp_size;

	*pp_block_out = NULL;
	*psz_block_out = 0;

	remain_out = LZPT_MAX_DECOMP_BLOCK_SIZE;
	if (!(p_block = malloc(remain_out))) {
		fprintf(stderr, "lzpt_decompress_block(): allocation error!\n");
		return 1;
	}

	remain_in = sz_block_in;
	p_in = p_block_in;
	p_out = p_block;
	total_decomp_size = 0;
	while((remain_in > LZPT_BLOCK_ALIGN_MASK) && (remain_out > 0)) {
		this_declen = lz77_inflate(p_in, remain_in, p_out, remain_out, &next_in);
		if (this_declen <= 0) {
			break;
		}
		p_out += this_declen;
		remain_out -= this_declen;
		total_decomp_size += this_declen;

		remain_in -= (next_in - p_in);
		p_in = next_in;
	}


	if ((remain_in > LZPT_BLOCK_ALIGN_MASK) || (remain_out < 0)) {
		fprintf(stderr, "lzpt_decompress_block(): decompress error!\n");
		free((void *)p_block);
		return 1;
	}

	*pp_block_out = p_block;
	*psz_block_out = total_decomp_size;

if (total_decomp_size != 65536) {
	printf("ERROR!\n");
	exit(1);
}
	return 0;
}


void
lzpt_free_toc(unsigned char *p_toc, size_t toc_nentries)
{
	if (p_toc) free((void *)p_toc);
}

